use crate::Link;
use crate::searchers::Searcher;

use reqwest;
use reqwest::StatusCode;

use select::document::Document;
use select::predicate::Name;

use retry::retry;
use retry::delay::Fixed;

use std;

use mopa::Any;

const NUM_RETRIES: usize = 10;
const RETRY_WAIT_MILLIS: u64 = 500;

#[derive(Debug)]
pub struct WikiPage {
    name: String,
    url: String
}

impl Link for WikiPage {
    fn label (&self) -> &str {
        &self.name
    }

    fn url (&self) -> &str {
        &self.url
    }

    fn as_any (&self) -> &dyn Any {
        self
    }
}

pub struct MediawikiSearcher {
    baseurl: String
}

impl MediawikiSearcher {
    pub fn new (url: String) -> MediawikiSearcher {
        MediawikiSearcher {
            baseurl: url
        }
    }

    fn do_search (&self, name: &str) -> Result<String, Error> {
        let api_url = &format!("{}{}", self.baseurl, name);
        let response = retry(Fixed::from_millis(RETRY_WAIT_MILLIS).take(NUM_RETRIES), || {
            reqwest::blocking::get(api_url)
        })?;

        match response.status() {
            StatusCode::OK => {
                match response.text() {
                    Ok(text) => Ok(text),
                    Err(error) => Result::Err(Error::Http(error))
                }
            },
            _ => {
                Result::Err(Error::Other(String::from("Not Found")))
            }
        }
    }

    fn parse_entry (&self, page: String) -> Result<WikiPage, Error> {
        let document = Document::from(&page[..]);
        let page_name = String::from(document.find(Name("h1")).into_iter().next().expect("expected h1").text());
        let page_url = format!("{}{}", self.baseurl, page_name.replace(" ", "_"));

        Result::Ok(WikiPage {
            name: page_name,
            url: page_url
        })
    }
}

impl Searcher<WikiPage> for MediawikiSearcher {
    fn exact_search (&self, name: &str) -> Option<WikiPage> {
        self.do_search(name).and_then(|content| self.parse_entry(content)).ok()
    }
}

#[derive(Debug)]
enum Error {
    Http(reqwest::Error),
    Io(std::io::Error),
    Other(String)
}

impl From<reqwest::Error> for Error {
    fn from (error: reqwest::Error) -> Error {
        Error::Http(error)
    }
}

impl From<std::io::Error> for Error {
    fn from (error: std::io::Error) -> Error {
        Error::Io(error)
    }
}

impl From<retry::Error<reqwest::Error>> for Error {
    fn from (err: retry::Error<reqwest::Error>) -> Error {
        match err {
            retry::Error::Operation { error, total_delay, tries } => {
                Error::Http(error)
            }
        }
    }
}
